import 'dart:convert';

import 'package:crypto/crypto.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:provider/provider.dart';
import 'package:retgoo_sdk/forms/forms.dart';
import 'package:retgoo_sdk/models/directform.dart';
import 'package:retgoo_sdk/pages/diagram_page.dart';
import 'package:retgoo_sdk/pages/not_implemented_page.dart';
import 'package:retgoo_sdk/pages/process_information_page.dart';
import 'package:retgoo_sdk/providers/process_provider.dart';
import 'package:retgoo_sdk/widgets/retgoo_server.dart';
import 'package:shrink_sidemenu/shrink_sidemenu.dart';

class ProcessPage extends StatefulWidget {
  final int idProcessInstance;
  final int idActivityNotification;

  ProcessPage({Key key, this.idActivityNotification, this.idProcessInstance})
      : super(key: key);

  @override
  _ProcessPageState createState() => _ProcessPageState();
}

class _ProcessPageState extends State<ProcessPage> {
  ProcessProvider provider;

  @override
  void initState() {
    super.initState();
    provider = ProcessProvider(
      idProcessInstance: widget.idProcessInstance,
      idNotificationActivity: widget.idActivityNotification,
    );
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    RetGooServer.of(context).setListener("update_client_token", (value) {
      provider.refreshToken();
    });

    RetGooServer.of(context).setListener("task_token_completed", (value) {
      provider.refreshToken();
    });

    RetGooServer.of(context).setListener("activity_token_list", (value) {
      provider.refreshActivities();
    });
  }

  @override
  deactivate() {
    RetGooServer.of(context).removeListener("task_token_completed");
    RetGooServer.of(context).removeListener("activity_token_list");
    RetGooServer.of(context).removeListener("update_client_token");
    super.deactivate();
  }

  Widget build(BuildContext context) {
    return ChangeNotifierProvider.value(
      value: provider,
      child: ProcessContent(),
    );
  }
}

class ProcessContent extends StatefulWidget {
  const ProcessContent({
    Key key,
  }) : super(key: key);

  @override
  _ProcessContentState createState() => _ProcessContentState();
}

class _ProcessContentState extends State<ProcessContent> {
  UserFormTask task;
  Offset offset;

  final menuKey = GlobalKey<SideMenuState>();
  int _currentIndex = 0;

  buildBody() {
    return AnnotatedRegion(
      value: SystemUiOverlayStyle.dark.copyWith(
        statusBarColor: Colors.transparent,
      ),
      child: SideMenu(
        key: menuKey,
        background: Theme.of(context).scaffoldBackgroundColor,
        radius: BorderRadius.circular(48.0),
        type: SideMenuType.slideNRotate,
        closeIcon: null,
        menu: buildMenu(),
        child: Stack(
          children: [
            Positioned.fill(
              child: IndexedStack(
                index: _currentIndex,
                children: [
                  buildForm(),
                  ProcessInformationPage(),
                  DiagramPage(),
                ],
              ),
            ),
            Positioned(
              top: offset?.dy ?? MediaQuery.of(context).size.height - 16 - 64,
              right: 0,
              child: SideButton(
                onMoved: (offset) {
                  setState(() {
                    this.offset = offset;
                  });
                },
                onTap: () {
                  menuKey.currentState.openSideMenu();
                },
              ),
            ),
          ],
        ),
      ),
    );
  }

  buildMenu() {
    return SafeArea(
      top: false,
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          Padding(
            padding: const EdgeInsets.symmetric(horizontal: 16.0),
            child: Text(
              task.taskName,
              style: TextStyle(
                fontSize: 24.0,
                fontWeight: FontWeight.w900,
              ),
            ),
          ),
          Divider(),
          ListTile(
            selected: _currentIndex == 0,
            title: Text("Form"),
            leading: Icon(CupertinoIcons.doc_on_clipboard),
            onTap: () {
              setState(() {
                _currentIndex = 0;
              });
              menuKey.currentState.closeSideMenu();
            },
          ),
          ListTile(
            selected: _currentIndex == 1,
            title: Text("Information"),
            leading: Icon(CupertinoIcons.info),
            onTap: () {
              setState(() {
                _currentIndex = 1;
              });
              menuKey.currentState.closeSideMenu();
            },
          ),
          ListTile(
            selected: _currentIndex == 2,
            title: Text("Diagram"),
            leading: Icon(CupertinoIcons.flowchart),
            onTap: () {
              setState(() {
                _currentIndex = 2;
              });
              menuKey.currentState.closeSideMenu();
            },
          ),
          Divider(),
          ListTile(
            title: Text("Close"),
            leading: Icon(CupertinoIcons.chevron_back),
            onTap: () {
              Navigator.pop(context);
            },
          ),
        ],
      ),
    );
  }

  var userTask;
  buildForm() {
    return Selector<ProcessProvider, ProcessState>(
      selector: (context, provider) => provider.state,
      builder: (context, state, child) {
        if (state == ProcessState.completed) {
          return Container(
            alignment: Alignment.center,
            child: Text("completed"),
          );
        }

        final directForm =
            Provider.of<ProcessProvider>(context, listen: false).directForm;
        final registrationId =
            encodeTaskId(directForm.idMasterProcess, task.taskId);

        if (!registeredUserTasks.containsKey(registrationId)) {
          return NotImplementedPage(directForm: directForm);
        }

        if (userTask == null) {
          userTask = registeredUserTasks[registrationId]();
        }

        return userTask.attach(context);
      },
    );
  }

  String encodeTaskId(int masterProcessID, String taskID) {
    final sha = sha256.newInstance();
    final digest = sha.convert(utf8.encode("$masterProcessID/$taskID"));
    return digest.toString().toUpperCase();
  }

  @override
  Widget build(BuildContext context) {
    return Selector<ProcessProvider, DirectForm>(
      selector: (context, provider) => provider.directForm,
      builder: (context, directForm, child) {
        if (directForm == null) {
          return Scaffold(
            appBar: AppBar(
              title: Text("Loading"),
            ),
            body: Center(
              child: CircularProgressIndicator(),
            ),
          );
        }

        task = directForm?.taskInfo?.mainForm;
        return Scaffold(
          body: buildBody(),
        );
      },
    );
  }
}

class SideButton extends StatefulWidget {
  final double height;
  final ValueChanged<Offset> onMoved;
  final GestureTapCallback onTap;
  const SideButton({
    Key key,
    this.height = 64,
    this.onMoved,
    this.onTap,
  }) : super(key: key);

  @override
  _SideButtonState createState() => _SideButtonState();
}

class _SideButtonState extends State<SideButton> with TickerProviderStateMixin {
  final buttonKey = GlobalKey();

  @override
  Widget build(BuildContext context) {
    return AnimatedOpacity(
      opacity: 0.5,
      duration: Duration(milliseconds: 200),
      child: Draggable(
        affinity: Axis.vertical,
        axis: Axis.vertical,
        childWhenDragging: Container(),
        onDragEnd: (details) {
          double y = details.offset.dy;
          if (y < 0) {
            y = 0;
          }

          if (y > MediaQuery.of(context).size.height - widget.height) {
            y = MediaQuery.of(context).size.height - widget.height;
          }

          final offset = Offset(details.offset.dx, y);
          if (widget.onMoved != null) {
            widget.onMoved(offset);
          }
        },
        feedback: buildSideMenu(),
        child: GestureDetector(
          onTap: widget.onTap,
          child: buildSideMenu(),
        ),
      ),
    );
  }

  Widget buildSideMenu() {
    return Container(
      key: buttonKey,
      height: widget.height,
      decoration: BoxDecoration(
        color: Theme.of(context).primaryColor,
        borderRadius: BorderRadius.only(
          topLeft: Radius.circular(100),
          bottomLeft: Radius.circular(100),
        ),
      ),
      padding: const EdgeInsets.all(4.0),
      child: Padding(
        padding: const EdgeInsets.all(8.0),
        child: Icon(
          CupertinoIcons.fullscreen_exit,
          color: Colors.white,
        ),
      ),
    );
  }
}
