import 'package:flutter/material.dart';
import 'package:retgoo_sdk/models/diagram_token.dart';
import 'package:retgoo_sdk/models/directform.dart';
import 'package:retgoo_sdk/models/process_activity.dart';
import 'package:retgoo_sdk/protocol/api.dart';
import 'package:retgoo_sdk/utils/xml_builder.dart';

enum ProcessState { initializing, ready, completed }

class ProcessProvider with ChangeNotifier {
  ProcessState state = ProcessState.initializing;

  final int idProcessInstance;
  final int idNotificationActivity;

  DirectForm _directForm;
  DirectForm get directForm => _directForm;

  ProcessProvider({this.idProcessInstance, this.idNotificationActivity}) {
    _getDirectForm();
    refreshActivities();
    refreshToken();
  }

  Future<void> _getDirectForm() async {
    var response = await api.post("process/getdirectformheader", data: {
      "IdActivityNotificationItem": idNotificationActivity,
      "IdProcessInstance": idProcessInstance,
    });

    if (response.statusCode != 200) {
      return;
    }

    response = await api.post("process/getdirectform", data: response.data);
    if (response.statusCode == 200) {
      _directForm = DirectForm.fromJson(response.data);
      state = ProcessState.ready;
      notifyListeners();
    }
  }

  List<ProcessActivity> _activities = [];
  List<ProcessActivity> get activities => _activities;

  refreshActivities() async {
    final response = await api.post(
      "process/getActivityList",
      data: idProcessInstance.toString(),
    );

    if (response.statusCode == 200) {
      _activities = (response.data as List)
          .map((json) => ProcessActivity.fromJson(json))
          .toList();

      notifyListeners();
    }
  }

  Map<String, ProcessToken> _tokens = {};
  Map<String, ProcessToken> get tokens => _tokens;
  refreshToken() async {
    final result = await api.post(
      "process/GetDiagramToken",
      data: idProcessInstance,
    );

    if (result.statusCode == 200) {
      Map<String, ProcessToken> _tokens = {};
      List list = result.data["Items"];

      List<ProcessToken> res =
          list.map((json) => ProcessToken.fromJson(json)).toList();

      for (var item in res) {
        if (_tokens.containsKey(item.id)) {
          if (_tokens[item.id].isCurrentToken) {
            continue;
          }
        }
        _tokens[item.id] = item;
      }

      this._tokens = _tokens;
      notifyListeners();
    }
  }

  Future<bool> claimProcess(String flowButtonId) async {
    String payload = _createClaimTaskRequest(
      directForm: directForm,
      flowButtonElementId: flowButtonId,
    );

    final response = await api.post(
      "process/claimprocess",
      data: payload,
    );

    if (response.statusCode == 200 && response.data["Success"]) {
      state = ProcessState.completed;
      notifyListeners();
      return true;
    }

    return false;
  }

  String _createClaimTaskRequest(
      {DirectForm directForm, String flowButtonElementId}) {
    try {
      XMLCreator creator = XMLCreator(optimizedNameSpaces: true);
      creator.addProcessing(
          addedAttribute: 'encoding="utf-8" standalone="yes"');

      creator.addElement("BPMNSocket", nest: () {
        creator.addElement("processInstance", nest: () {
          creator.addAttribute("id", directForm.idProcessInstance.toString());
          creator.addAttribute("command", BPMNSocketType.claimTask.toString());
          creator.addAttribute("idDataTransmit", "df0c59");
          creator.addElement("resourceClaim", nest: () {
            //creator.addAttribute("userId", currentUser.id.toString());
            creator.addAttribute("taskId", directForm.taskInfo.mainForm.taskId);
            creator.addAttribute("portalType", "10");
            creator.addAttribute("controlId", flowButtonElementId);
          });

          creator.addElement("dataInputValueList", nest: () {
            if (directForm.variable.values == null ||
                directForm.variable.values.length <= 0) {
              return;
            }

            for (final variable in directForm.variable.values) {
              creator.addElement("dataInputValue", nest: () {
                creator.addAttribute("id", variable.id);
                creator.addAttribute("typeId", variable.type);

                var cDataValue = variable.rawValue;
                // if (int.tryParse(variable.type) == BPMNDataType.integer) {
                //   if (variable.rawValue is double) {
                //     cDataValue = (variable.rawValue as double).toInt();
                //   }
                // }

                creator.addElement("value", nest: () {
                  creator.addCData(cDataValue?.toString() ?? "");
                });
              });
            }
          });

          creator.addElement("formCoachList", nest: () {
            creator.addElement("formCoachItem", nest: () {
              creator.addAttribute(
                  "taskId", directForm.taskInfo.mainForm.taskId);
            });
          });
        });
      });

      return creator.finish(prettyXML: true);
    } catch (ex) {
      ex.toString();
    }
    return null;
  }
}

class BPMNSocketType {
  static final int none = 0;

  static final int elementClick = 1;
  static final int expressionClick = 2;

  static final int updateClientToken = 20;
  static final int updateTaskPaused = 21;
  static final int requestUpdateClientToken = 22;
  static final int taskTokenCompleted = 23;
  static final int requestActivityTokenList = 24;
  static final int activityTokenList = 25;

  static final int loadInstanceProcess = 50;
  static final int processInstanceCompleted = 51;
  static final int requestLoadInstanceProcess = 52;
  static final int requestCreateInstanceProcess = 53;
  static final int diagramProcessInstance = 54;
  static final int requestSendDiagramProcessInstance = 55;
  static final int processTabOpen = 56;
  static final int processTabClose = 57;
  static final int processTabSelected = 58;

  static final int formCoachXML = 70;
  static final int dataSetFormCoachXML = 71;
  static final int claimTask = 72;
  static final int requestSendFormCoachXML = 73;
  static final int formCoachXMLRestricted = 74;
  static final int replayClaimTask = 75;
  static final int userFormOpened = 76;
  static final int userFormClosed = 77;
  static final int formLocked = 78;
  static final int userFormMessage = 79;
  static final int userFormReady = 80;
  static final int userFormActivity = 81;
  static final int userFormRenderAll = 82;

  static final int notificationActivity = 90;
  static final int requestNotificationActivityList = 91;
  static final int notificationActivityList = 92;
  static final int notificationUserUsed = 93;
  static final int requestNotificationNumber = 94;
  static final int notificationNumber = 95;
  static final int notificationUserOnline = 96;
  static final int notificationActivityDirectForm = 97;
  static final int notificationRemoveById = 98;
  static final int notificationActivityDirectFormNew = 99;

  static final int userLogin = 110;
  static final int userKick = 111;

  static final int requestDataVariableSchemeValue = 120;
  static final int requestDataVariableValue = 121;
  static final int dataVariableSchemeValue = 122;
  static final int requestSendDataVariableFormCoachXML = 123;
  static final int dataVariableFormCoachXML = 124;

  static final int pingSend = 500;
  static final int pingReplay = 501;

  static final int welcomeUserLogin = 1000;
  static final int serverMessage = 1001;
  static final int serverConnectedInfo = 1002;
  static final int autoUpdateStudio = 1003;
  static final int autoUpdatePortal = 1004;
  static final int autoUpdateBI = 1005;

  static final int updateModelExe = 5000;
}

class BPMNDataType {
  static final int boolean = 0;
  static final int byte = 1;
  static final int char = 2;
  static final int date = 3;
  static final int decimal = 4;
  static final int double = 5;
  static final int integer = 6;
  static final int long = 7;
  static final int sByte = 8;
  static final int short = 9;
  static final int single = 10;
  static final int string = 11;
  static final int uInteger = 12;
  static final int uLong = 13;
  static final int uShort = 14;
  static final int time = 15;
  static final int dateTime = 16;
  static final int timeSpan = 17;
  static final int excel = 18;
  static final int image = 19;
  static final int binary = 20;
  static final int none = 100;
}
