import 'dart:math' as math;
import 'dart:ui';

import 'package:flutter/material.dart';
import 'package:retgoo_sdk/models/diagram_token.dart';
import 'package:retgoo_sdk/models/directform.dart';
import 'package:xml/xml.dart' as xml;

class BpmnDiagram extends StatelessWidget {
  final DirectForm directForm;
  final Map<String, ProcessToken> tokens;

  BpmnDiagram({this.directForm, this.tokens = const {}});
  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
      builder: (context, constraint) {
        final size = _getDiagramSize();
        final width = constraint.maxWidth;
        final ratio = width / size.width;

        return InteractiveViewer(
          boundaryMargin: const EdgeInsets.all(16.0),
          maxScale: 5.0,
          minScale: 0.2,
          panEnabled: true,
          scaleEnabled: true,
          constrained: true,
          alignPanAxis: true,
          child: CustomPaint(
            size: Size(size.width * ratio, size.height * ratio),
            isComplex: true,
            painter: _BpmnDiagramPainter.fromXmlString(
              directForm.diagram.bpmn20Export,
              fontFamily: Theme.of(context).textTheme.bodyText1.fontFamily,
              scale: ratio,
              tokenList: tokens,
            ),
          ),
        );
      },
    );
  }

  Size _getDiagramSize() {
    xml.XmlDocument bpmn20Export =
        xml.XmlDocument.parse(directForm.diagram.bpmn20Export);
    var diagramItems = bpmn20Export.findAllElements("IDiagramItem").toList();

    Size size = Size.zero;
    for (var diagramItem in diagramItems) {
      if (diagramItem.getAttribute("isMainDiagram") == "true") {
        double w = 0;
        double h = 0;

        var nodes = diagramItem.findAllElements("MxtNode").toList();
        for (var lane in nodes) {
          double height = double.parse(lane.getAttribute("boundsHeight"));
          double width = double.parse(lane.getAttribute("boundsWidth"));
          double x = double.parse(lane.getAttribute("boundsX"));
          double y = double.parse(lane.getAttribute("boundsY"));

          if (x + width > size.width) {
            w = x + width;
          }

          if (y + height > size.height) {
            h = y + height;
          }

          if (w * h > size.width * size.height) {
            size = Size(w, h);
          }
        }

        break;
      }
    }

    return Size(size.width + 20, size.height);
  }
}

class _BpmnDiagramPainter extends CustomPainter {
  final Map<String, ProcessToken> tokenList;
  final String xmlString;

  final List<xml.XmlElement> lanes;
  final List<xml.XmlElement> nodes;
  final List<xml.XmlElement> links;
  final double scale;
  final String fontFamily;
  final Color completedColor;
  final Color currentColor;

  _BpmnDiagramPainter({
    this.lanes,
    this.nodes,
    this.links,
    this.scale = 1.0,
    this.fontFamily,
    this.tokenList = const {},
    this.completedColor = Colors.purple,
    this.currentColor = Colors.blue,
    this.xmlString,
  });

  factory _BpmnDiagramPainter.fromXmlString(
    String xmlString, {
    double scale,
    String fontFamily,
    Map<String, ProcessToken> tokenList,
  }) {
    xml.XmlDocument bpmn20Export = xml.XmlDocument.parse(xmlString);

    var diagramItems = bpmn20Export.findAllElements("IDiagramItem").toList();

    List<xml.XmlElement> lanes;
    List<xml.XmlElement> nodes;
    List<xml.XmlElement> links;

    diagramItems.forEach((diagramItem) {
      if (diagramItem.getAttribute("isMainDiagram") == "true") {
        lanes = diagramItem.findElements("MxtLane").toList();
        nodes = diagramItem.findElements("MxtNode").toList();
        links = diagramItem.findElements("MxtLink").toList();
      }
    });

    return _BpmnDiagramPainter(
      lanes: lanes,
      nodes: nodes,
      links: links,
      scale: scale,
      fontFamily: fontFamily,
      tokenList: tokenList,
      xmlString: xmlString,
    );
  }

  double width;
  double height;

  @override
  void paint(Canvas canvas, Size size) {
    double x = size.width;
    double y = size.height;
    width = size.width;
    height = 0;

    try {
      lanes.forEach((lane) {
        Bound bound = _getBound(lane);
        bool isHorizontal = lane.getAttribute("isHorizontal") == "true";

        x = math.min<double>(bound.x, x);

        if (isHorizontal) {
          y = math.min<double>(bound.y, y);
        }

        width = math.max<double>(bound.width, width);
        height = math.max<double>(bound.height, height);
      });

      Paint line = new Paint()
        ..isAntiAlias = true
        ..color = Colors.white
        ..strokeCap = StrokeCap.round
        ..style = PaintingStyle.fill
        ..strokeWidth = 0.5 * scale;

      canvas.drawRRect(
          RRect.fromRectAndRadius(
            Rect.fromLTWH(x, y, width, height),
            Radius.circular(4 * scale),
          ),
          line);

      line = new Paint()
        ..isAntiAlias = true
        ..color = Colors.grey.shade300
        ..strokeCap = StrokeCap.round
        ..style = PaintingStyle.stroke
        ..strokeWidth = 0.5 * scale;

      canvas.drawRRect(
          RRect.fromRectAndRadius(
            Rect.fromLTWH(x, y, width, height),
            Radius.circular(4 * scale),
          ),
          line);

      lanes.forEach((lane) {
        _drawLane(canvas, size, lane);
      });

      nodes.forEach((node) {
        _drawNode(canvas, size, node);
      });

      links.forEach((link) {
        _drawLink(canvas, size, link);
      });
    } on Exception catch (e) {
      print(e);
    }
  }

  _getBound(xml.XmlElement element) {
    double height = double.parse(element.getAttribute("boundsHeight"));
    double width = double.parse(element.getAttribute("boundsWidth"));
    double x = double.parse(element.getAttribute("boundsX"));
    double y = double.parse(element.getAttribute("boundsY"));

    return Bound(
      height: height * scale,
      width: width * scale,
      x: x * scale,
      y: y * scale,
    );
  }

  void drawText(Canvas canvas, String text, Color color, Bound bound,
      double fontSize, double angleRotationInRadians) {
    canvas.save();

    TextSpan span = TextSpan(
        style:
            TextStyle(color: color, fontSize: fontSize, fontFamily: fontFamily),
        text: text);

    TextPainter tp = TextPainter(
      text: span,
      textAlign: TextAlign.center,
      textDirection: TextDirection.ltr,
      textScaleFactor: 1 * scale,
    );

    tp.layout(
        maxWidth: angleRotationInRadians != 0 ? bound.height : bound.width);

    canvas.translate(bound.x + (bound.width / 2), bound.y + (bound.height / 2));
    canvas.rotate(angleRotationInRadians);
    tp.paint(canvas, Offset(-(tp.width / 2), -(tp.height / 2)));
    canvas.restore();
  }

  _drawLane(Canvas canvas, Size size, xml.XmlElement lane,
      {int laneCountInParent = 0, int currentChildIndex = -1}) {
    bool isHorizontal = lane.getAttribute("isHorizontal") == "true";
    String title = lane.getAttribute("title");
    Bound bound = _getBound(lane);

    xml.XmlElement parent = lane.parent;
    bool isProcessLane = parent.name.local == "IDiagramItem";

    Paint line = new Paint()
      ..isAntiAlias = true
      ..color = Colors.deepOrange
      ..strokeCap = StrokeCap.round
      ..style = PaintingStyle.fill
      ..strokeWidth = 0.5 * scale;

    if (isHorizontal && !isProcessLane) {
      canvas.drawRRect(
          RRect.fromRectAndCorners(
            Rect.fromLTWH(
                bound.x + (2 * scale),
                currentChildIndex == 0
                    ? bound.y - (4 * scale)
                    : bound.y + (0.5 * scale),
                bound.width - (4 * scale),
                currentChildIndex == 0
                    ? bound.height + ((laneCountInParent == 1 ? 8 : 4) * scale)
                    : (currentChildIndex == laneCountInParent - 1
                        ? bound.height +
                            ((laneCountInParent == 1 ? 8 : 4) * scale)
                        : bound.height - (scale))),
            topLeft: currentChildIndex == 0
                ? Radius.circular(2 * scale)
                : Radius.circular(0),
            topRight: currentChildIndex == 0
                ? Radius.circular(2 * scale)
                : Radius.circular(0),
            bottomLeft: currentChildIndex == laneCountInParent - 1
                ? Radius.circular(2 * scale)
                : Radius.circular(0),
            bottomRight: currentChildIndex == laneCountInParent - 1
                ? Radius.circular(2 * scale)
                : Radius.circular(0),
          ),
          line);

      if (laneCountInParent > 0 && currentChildIndex < laneCountInParent - 1) {
        Path path = Path();
        path.moveTo(bound.x + bound.width, bound.y + bound.height);
        path.lineTo(width, bound.y + bound.height);
        line.style = PaintingStyle.stroke;
        line.color = Colors.grey.shade300;
        line.strokeWidth = 0.5 * scale;
        canvas.drawPath(path, line);
      }
    }

    if (title.isNotEmpty) {
      if (isHorizontal) {
        drawText(
            canvas,
            title,
            isProcessLane ? Colors.grey.shade900 : Colors.white,
            bound,
            isProcessLane ? 5 : 3.5,
            -1.570800000709);
      }
    }

    var childLanes = lane.findElements("MxtLane").toList();
    if (childLanes != null && childLanes.length > 0) {
      childLanes.forEach((child) {
        _drawLane(canvas, size, child,
            laneCountInParent: childLanes.length,
            currentChildIndex: childLanes.indexOf(child));
      });
    }
  }

  _drawNode(Canvas canvas, Size size, xml.XmlElement node) {
    String elementType = node.getAttribute("elementType");
    String idElement = node.getAttribute("idElement");

    bool isCompleted = false;
    bool isCurrentToken = false;
    if (tokenList.containsKey(idElement)) {
      isCompleted = true;
      isCurrentToken = tokenList[idElement].isCurrentToken;
    }

    Bound bound = _getBound(node);

    Paint line = new Paint()
      ..color = Colors.grey.shade200
      ..strokeCap = StrokeCap.round
      ..style = PaintingStyle.stroke
      ..strokeWidth = 0.5 * scale;

    Paint fill = new Paint()
      ..color = Colors.black
      ..strokeCap = StrokeCap.round
      ..style = PaintingStyle.fill
      ..strokeWidth = 0.2 * scale;

    switch (elementType) {
      case "startEvent":
        fill.color = isCompleted ? completedColor : Colors.lightGreen;
        canvas.drawOval(
            Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height), fill);

        canvas.drawOval(
            Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height), line);

        xml.XmlElement label = node.findElements("MxtLabel").first;
        if (label != null) {
          String text = label.getAttribute("text");
          if (text.isNotEmpty) {
            drawText(
                canvas, text, Colors.grey.shade700, _getBound(label), 2, 0);
          }
        }
        break;

      case "endEvent":
        fill.color = isCompleted ? completedColor : Colors.red[300];
        canvas.drawOval(
            Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height), fill);

        canvas.drawOval(
            Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height), line);

        xml.XmlElement label = node.findElements("MxtLabel").first;
        if (label != null) {
          String text = label.getAttribute("text");
          if (text.isNotEmpty) {
            drawText(
                canvas, text, Colors.grey.shade700, _getBound(label), 2, 0);
          }
        }
        break;

      case "exclusiveGateway":
        fill.color = isCompleted ? completedColor : Colors.yellow.shade300;
        Path path = Path()
          ..moveTo(bound.x + (bound.width / 2), bound.y)
          ..lineTo(bound.x, bound.y + (bound.height / 2))
          ..lineTo(bound.x + (bound.width / 2), bound.y + bound.height)
          ..lineTo(bound.x + bound.width, bound.y + (bound.height / 2))
          ..close();

        canvas.drawPath(path, fill);

        line.color = Colors.yellow;
        canvas.drawPath(path, line);

        xml.XmlElement label = node.findElements("MxtLabel").first;
        if (label != null) {
          String text = label.getAttribute("text");
          if (text.isNotEmpty) {
            drawText(
                canvas, text, Colors.grey.shade700, _getBound(label), 2, 0);
          }
        }
        break;

      case "userTask":
        fill.color = isCompleted
            ? (isCurrentToken ? currentColor : completedColor)
            : Colors.white;

        line.color = isCompleted
            ? (isCurrentToken ? currentColor : completedColor)
            : Colors.blue;

        canvas.drawRRect(
            RRect.fromRectAndRadius(
                Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height),
                Radius.circular(3 * scale)),
            fill);

        canvas.drawRRect(
            RRect.fromRectAndRadius(
                Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height),
                Radius.circular(3 * scale)),
            line);

        Color textColor =
            isCompleted || isCurrentToken ? Colors.white : Colors.grey.shade800;

        drawText(canvas, node.getAttribute("name"), textColor, bound, 2, 0);

        break;

      case "intermediateThrowEvent;Meitan.BPM.bpmn.model.core.common.events.MessageEventDefinition":
        fill.color = isCompleted ? completedColor : Colors.orange;
        canvas.drawOval(
            Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height), fill);

        canvas.drawOval(
            Rect.fromLTWH(bound.x, bound.y, bound.width, bound.height), line);

        xml.XmlElement label = node.findElements("MxtLabel").first;
        if (label != null) {
          String text = label.getAttribute("text");
          if (text.isNotEmpty) {
            drawText(
                canvas, text, Colors.grey.shade700, _getBound(label), 2, 0);
          }
        }
        break;
      default:
        print(elementType);
    }
  }

  _drawLink(Canvas canvas, Size size, xml.XmlElement link) {
    String idElement = link.getAttribute("idElement");
    bool isCompleted = false;
    bool isCurrentToken = false;
    if (tokenList.containsKey(idElement)) {
      isCompleted = true;
      isCurrentToken = tokenList[idElement].isCurrentToken;
    }

    var waypoints = link.findElements("waypoint").toList();
    if (waypoints != null && waypoints.isNotEmpty) {
      Paint line = new Paint()
        ..color = isCompleted
            ? (isCurrentToken ? currentColor : completedColor)
            : Colors.amber.shade300
        ..strokeCap = StrokeCap.round
        ..strokeJoin = StrokeJoin.round
        ..style = PaintingStyle.stroke
        ..isAntiAlias = true
        ..strokeWidth = 0.5 * scale;

      Path path = Path();
      path.moveTo(double.parse(waypoints[0].getAttribute("x")) * scale,
          double.parse(waypoints[0].getAttribute("y")) * scale);

      for (int i = 0; i < waypoints.length - 1; i++) {
        var waypoint = waypoints[i + 1];
        double x1 = double.parse(waypoint.getAttribute("x")) * scale;
        double y1 = double.parse(waypoint.getAttribute("y")) * scale;
        path.lineTo(x1, y1);
      }

      canvas.drawPath(path, line);

      if (waypoints.length > 2) {
        var waypoint1 = waypoints[waypoints.length - 2];
        var waypoint2 = waypoints[waypoints.length - 1];

        double wp1x = double.parse(waypoint1.getAttribute("x")) * scale;
        double wp1y = double.parse(waypoint1.getAttribute("y")) * scale;

        double wp2x = double.parse(waypoint2.getAttribute("x")) * scale;
        double wp2y = double.parse(waypoint2.getAttribute("y")) * scale;

        if (wp1x == wp2x && wp1y == wp2y) {
          waypoint1 = waypoints[waypoints.length - 3];
          waypoint2 = waypoints[waypoints.length - 2];

          wp1x = double.parse(waypoint1.getAttribute("x")) * scale;
          wp1y = double.parse(waypoint1.getAttribute("y")) * scale;

          wp2x = double.parse(waypoint2.getAttribute("x")) * scale;
          wp2y = double.parse(waypoint2.getAttribute("y")) * scale;
        }

        var direction = "right";
        if (wp1x == wp2x) {
          if (wp1y < wp2y) {
            direction = "bottom";
          } else if (wp1y > wp2y) {
            direction = "top";
          }
        } else if (wp1y == wp2y) {
          if (wp1x < wp2x) {
            direction = "right";
          } else if (wp1x > wp2x) {
            direction = "left";
          }
        }

        switch (direction) {
          case "right":
            path = Path();
            path.moveTo(wp2x - (2.5 * scale), wp2y);
            path.lineTo(wp2x - (2.5 * scale), wp2y - (1.5 * scale));
            path.lineTo(wp2x, wp2y);
            path.lineTo(wp2x - (2.5 * scale), wp2y + (1.5 * scale));
            path.lineTo(wp2x - (2.5 * scale), wp2y);
            path.close();

            line.style = PaintingStyle.fill;
            canvas.drawPath(path, line);
            break;
          case "left":
            path = Path();
            path.moveTo(wp2x + (2.5 * scale), wp2y);
            path.lineTo(wp2x + (2.5 * scale), wp2y - (1.5 * scale));
            path.lineTo(wp2x, wp2y);
            path.lineTo(wp2x + (2.5 * scale), wp2y + (1.5 * scale));
            path.lineTo(wp2x + (2.5 * scale), wp2y);
            path.close();

            line.style = PaintingStyle.fill;
            canvas.drawPath(path, line);
            break;
          case "top":
            path = Path();
            path.moveTo(wp2x, wp2y + (2.5 * scale));
            path.lineTo(wp2x - (1.5 * scale), wp2y + (2.5 * scale));
            path.lineTo(wp2x, wp2y);
            path.lineTo(wp2x + (1.5 * scale), wp2y + (2.5 * scale));
            path.lineTo(wp2x, wp2y + (2.5 * scale));
            path.close();

            line.style = PaintingStyle.fill;
            canvas.drawPath(path, line);
            break;
          case "bottom":
            path = Path();
            path.moveTo(wp2x, wp2y - (2.5 * scale));
            path.lineTo(wp2x - (1.5 * scale), wp2y - (2.5 * scale));
            path.lineTo(wp2x, wp2y);
            path.lineTo(wp2x + (1.5 * scale), wp2y - (2.5 * scale));
            path.lineTo(wp2x, wp2y - (2.5 * scale));
            path.close();

            line.style = PaintingStyle.fill;
            canvas.drawPath(path, line);
            break;
        }

        xml.XmlElement label = link.findElements("MxtLabel").first;
        if (label != null) {
          String text = label.getAttribute("text");
          if (text.isNotEmpty) {
            drawText(
                canvas, text, Colors.grey.shade700, _getBound(label), 2, 0);
          }
        }
      }
    }
  }

  @override
  bool shouldRepaint(_BpmnDiagramPainter oldDelegate) =>
      xmlString != oldDelegate.xmlString || tokenList != oldDelegate.tokenList;
}

class Bound {
  double x;
  double y;
  double width;
  double height;
  Bound({this.x, this.y, this.width, this.height});
}
